package minishell;
import java.util.Hashtable;
import java.util.Vector;
import java.io.PrintWriter;

public class ClassPathLoader extends ClassLoader
{
    public static boolean 	cache = false;
    public static boolean 	debug = false;
    private ClassPathReader	reader;
    private static Hashtable	shared = new Hashtable();
    private static Vector 	sharedprefixes = new Vector();
    private static Vector 	privatenames = new Vector();
    private Hashtable		resolved;
    private int	depth = 0;

    void log(String msg)
    {
	if (!debug)
	    return;
	StringBuffer sb = new StringBuffer();
	for (int i = 0; i < depth; i++)
	    sb.append("  ");
	System.out.println(sb.toString() + "ClassPathLoader: " + msg);
    }

    Class shareSystemClass(String name) throws ClassNotFoundException
    {
	Class c;
	shared.put(name, c = findSystemClass(name));
	resolveClass(c);
	return c;
    }

    public static void listPrivateNames(PrintWriter p)
    {
	for (int i = 0; i < privatenames.size(); i++)
	    p.println(privatenames.elementAt(i));
    }

    public static void addPrivateName(String user_name)
    {
	privatenames.addElement(user_name);
    }

    public static void listSharedPrefixes(PrintWriter p)
    {
	for (int i = 0; i < sharedprefixes.size(); i++)
	    p.println(sharedprefixes.elementAt(i));
    }

    public static void addSharedPrefix(String user_name)
    {
	sharedprefixes.addElement(user_name);
    }

    /**
     * load a class
     */
    public Class loadClass(String user_name, boolean resolve) 
	throws ClassNotFoundException
    {
	String name;

	name = new String(user_name);

	depth++;
	log("Loading " + name + "/" + resolve);

	/*
	 * are we already sharing this class
	 */
	Class c = (Class)shared.get(name);
	if (c != null) {
	    log("Sharing " + name);
	    depth--;
	    return c;
	}

	c = (Class)resolved.get(name);
	if (c != null) {
	    log("Already resolved " + name);
	    depth--;
	    return c;
	}

	boolean priv = false;
	// see if the prefix matches one of the private names
	for (int i = 0; i < privatenames.size(); i++)
	    if (name.equals((String)privatenames.elementAt(i))) {
		log("Private copy for " + name);
		priv = true;
		break;
	    }

	// else see if the prefix matches one of the shared prefixes
	if (!priv) {
	    for (int i = 0; i < sharedprefixes.size(); i++) {
		if (name.startsWith((String)sharedprefixes.elementAt(i))) {
		    c = shareSystemClass(name);
		    if (c != null) {
			log("Sharing prefixed " + name);
			depth--;
			return c;
		    }
		}
	    }
	}

	// else use our byte code reader
	byte[]code = null;
	try {
	    code = reader.readByteCode(name);
	} catch (Exception e) {
	    log("couldn't read " + name + " " + e);
	    throw new ClassNotFoundException(name);
	}

	log("Defining " + name + " clen=" + code.length);

	c = defineClass(name, code, 0, code.length);
	if (resolve)
	    resolveClass(c);

	log("Loaded " + name);
	depth--;
	resolved.put(name, c);
	return c;
    }

    public static ClassPathLoader createClassLoader() 
	throws Exception
    {
	return createClassLoader(System.getProperty("java.class.path"));
    }

    public static ClassPathLoader createClassLoader(String classpath) 
	throws Exception
    {
	// copy classpath in
	ClassPathLoader cl = new ClassPathLoader(
		new ClassPathReader(cache, classpath));

	return cl;
    }

    private ClassPathLoader(ClassPathReader reader)
    {
	this.reader = reader;
	resolved = new Hashtable();
    }
}
